#!/usr/bin/perl 
use warnings;
use strict;

if ($#ARGV != 0) {
    print "This program takes exactly one command line argument.\n";
    exit;
}

## Set default values of parameters ##
my @viewpt = (0,0,0);
my @orig = (0,0,0);
my @vert = (0,0,1);
my $magnification = 1;
my $stepback = 1;
my @translation = (0,0);

## Initialize the input and output files ##
my $argname = $ARGV[0];
my $infilename = $argname . '.tdi';
my $outfilename = $argname . '.tdo';
# print "argname is $argname\n";
# exit;

# my $outfilename = $argname . '.tdo';

open(my $infile, "<$infilename") or die "Could not open $infilename.\n";
open(my $outfile, ">$outfilename") or die "Could not open $outfilename.\n";

## Do the actual processing ##
my $lineno = 0;
while (<$infile>) {
    $lineno++;
    my $line = $_;
    process_line($line);
}

## Clean up ##
close($infile) or die "Could not close $infilename.\n";
close($outfile) or die "Could not close $outfilename.\n";

#################
## Subroutines ##
#################

## This is the master subroutine ##
sub process_line {
    my $line = $_[0];
    chomp($line);
    if ($line =~ m{PRJT_set_viewpoint}) {
	setviewpoint($line);
    } elsif ($line =~ m{PRJT_set_origin}) {
	setorigin($line);
    } elsif ($line =~ m{PRJT_set_vertical}) {
	setvertical($line);
    } elsif ($line =~ m{PRJT_set_magnification}) {
	setmagnification($line);
    } elsif ($line =~ m{PRJT_set_stepback}) {
	setstepback($line);
    } elsif ($line =~ m{PRJT_set_translation}) {
	settranslation($line);
    } else {
	extract_points($line);
    }
}

## The next group of routines set parameters ##
sub setviewpoint {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    my @third = split(/,/, $second[0]);
    for (my $i = 0; $i < 3; $i++) { 
	$viewpt[$i] = $third[$i]; 
    }
}

sub setorigin {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    my @third = split(/,/, $second[0]);
    for (my $i = 0; $i < 3; $i++) { 
	$orig[$i] = $third[$i]; 
    }
}

sub setvertical {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    my @third = split(/,/, $second[0]);
    for (my $i = 0; $i < 3; $i++) { 
	$vert[$i] = $third[$i]; 
    }
}

sub setmagnification {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    $magnification = $second[0];
}

sub setstepback {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    $stepback = $second[0];
}

sub settranslation {
    my $line = $_[0];
    my @first = split(/\(/, $line);
    my @second = split(/\)/, $first[1]);
    my @third = split(/,/, $second[0]);
    for (my $i = 0; $i < 2; $i++) { 
	$translation[$i] = $third[$i]; 
    }
}

## The following routine extracts points and prints 2D realizations ##
sub extract_points {
    my $line = $_[0];
    if (length($line) == 0) {
	print $outfile "\n";
    } else {
	my @first = split(/%</, $line);
	my $i = 0;
	if ($line !~ m{/^%</}) {
	    $i++;
	    print $outfile "$first[0]";
	}
	while ($i <= $#first) {
	    my @second = split(/>%/, $first[$i]);
	    compute_projection($second[0]);
	    if (0 < $#second) {
		print $outfile "$second[1]";
	    }
	    $i++;
	}
    }
    print $outfile "\n";
}

## Finally we get to the actual numerical computation! ##
sub compute_projection {
    ## put the viewpoint in the relative coordinate system ##
    my @rawvee = vect_diff(\@viewpt,\@orig);
    my @vee = scalar_product($stepback,\@rawvee);

    ## compute the projected vertical vector ##
    my $beta = 1/(1 - inner_product(\@vee,\@vert)/inner_product(\@vee,\@vee));
    my @rawvert = affine_comb($beta,\@vee,\@vert);

    ## the projected horizontal vector is ##
    my @rawcross = cross_product(\@rawvert,\@vee);

    ## the associated unit vectors are ##
    my @prjtvert = scalar_product(1/norm(\@rawvert),\@rawvert);
    my @prjthori = scalar_product(1/norm(\@rawcross),\@rawcross);

    ## compute the projection of the input vector ##
    my @invec = split(/,/, $_[0]);
    my @eks = vect_diff(\@invec,\@orig);
    $beta = 1/(1 - inner_product(\@vee,\@eks)/inner_product(\@vee,\@vee));
    my @rawprjt = affine_comb($beta,\@vee,\@eks);
    
    ## find the coordinates in the (prjthori,prjtvert) coordinate system ##
    my $compone = inner_product(\@rawprjt,\@prjthori) * $magnification + $translation[0];
    my $comptwo = inner_product(\@rawprjt,\@prjtvert) * $magnification + $translation[1];

    ## output ##
    print $outfile "(";
    print $outfile sprintf("%.4f",$compone);
    print $outfile ",";
    print $outfile sprintf("%.4f",$comptwo);
    print $outfile ")";
}

## there are a number of standard vector operations ##
sub inner_product {
    my @vecone = @{$_[0]};
    my @vectwo = @{$_[1]};
    my $ip = 0;
    for (my $i = 0; $i <= 2; $i++) {
	$ip += $vecone[$i] * $vectwo[$i];
    }
    $ip;
}

sub norm {
    my @vecone = @{$_[0]};
    sqrt(inner_product(\@vecone,\@vecone));
}

sub vect_diff {
    my @vecone = @{$_[0]};
    my @vectwo = @{$_[1]};
    my @diff = (0,0,0);
    for (my $i = 0; $i <= 2; $i++) {
	$diff[$i] = $vecone[$i] - $vectwo[$i];
    }
    @diff;
}

sub vect_sum {
    my @vecone = @{$_[0]};
    my @vectwo = @{$_[1]};
    my @sum = (0,0,0);
    for (my $i = 0; $i <= 2; $i++) {
	$sum[$i] = $vecone[$i] + $vectwo[$i];
    }
    @sum;
}

sub scalar_product {
    my $scalar = $_[0];
    my @vector = @{$_[1]};
    my @prod = (0,0,0);
    for (my $i = 0; $i <= 2; $i++) {
	$prod[$i] = $scalar * $vector[$i];
    }
    @prod;
}

sub affine_comb {
    my $t = $_[0];
    my @x_zero = @{$_[1]};
    my @x_one = @{$_[2]};
    my @prod_zero = scalar_product(1 - $t,\@x_zero);
    my @prod_one = scalar_product($t,\@x_one);
    vect_sum(\@prod_zero,\@prod_one);
}

sub cross_product {
    my @vecone = @{$_[0]};
    my @vectwo = @{$_[1]};
    my @cross = ($vecone[1] * $vectwo[2] - $vecone[2] * $vectwo[1],
		 $vecone[2] * $vectwo[0] - $vecone[0] * $vectwo[2],
		 $vecone[0] * $vectwo[1] - $vecone[1] * $vectwo[0]);
    @cross;
}
## End of the line ##





